//=============================================================================
//
//   File : kvi_kvs_coresimplecommands_mr.cpp
//   Created on Fri 31 Oct 2003 00:04:25 by Szymon Stefanek
//
//   This file is part of the KVIrc IRC client distribution
//   Copyright (C) 2003 Szymon Stefanek <pragma at kvirc dot net>
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#define __KVIRC__

#include "kvi_kvs_coresimplecommands.h"

#include "kvi_window.h"
#include "kvi_out.h"
#include "kvi_locale.h"
#include "kvi_app.h"
#include "kvi_options.h"
#include "kvi_fileutils.h"
#include "kvi_filedialog.h"

#include "kvi_kvs_variantlist.h"
#include "kvi_kvs_script.h"

namespace KviKvsCoreSimpleCommands
{
	/*
		@doc: parse
		@type:
			command
		@title:
			parse
		@syntax:
			parse [-q] [-e] [-f] [-r] <filename> [parameters]
		@short:
			Executes commands from a file
		@switches:
			!sw: -e
			Causes the command to terminate with an error when the file
			can't be loaded: the error termination stops the parsing
			of the current script too. If the -e switch is not used then
			file load failures are eventually reported only as a warning
			or not reported at all (if -q is used).
			!sw: -f
			Causes the command to terminate with an error when the execution
			of the script loaded from disk generates an error.
			This error termination stops the parsing of the current script too.
			If the -f switch is not used then the loaded script errors
			do not stop the current script.
			!sw: -q
			Causes the command to run quietly without printing warnings about
			empty <filename> and file loading errors.
			!sw: -r
			Causes the return value of the script to be propagated to the
			calling context. This allows the usage of ${ } trick to extract
			this return value. See the examples section for a sample usage.
			
		@description:
			Executes commands from the external file <filename>.[br]
			<filename> can be an absolute or relative path.[br]
			[parameters] is a space separated string of parameters to be
			passed to the script. The parameter $0 will contain the
			name of the file being parsed, the other parameters will
			be available starting from $1.[br]
			If <filename> is an existing directory name a file dialog 
			will pop up, allowing you to choose a file to parse.
			The file dialog will start in the directory specified by <filename>.
			If you don't know the name of an existing directory, just pass "."
			or "/". The command just prints a warning if the file cannot be loaded for
			some reason. If you want the command to stop with an error in
			this case then use the -e switch.[br]
		@examples:
			[example]
				parse /home/pragma/myscript.kvs
			[/example]
			[example]
				[comment]# Here we are assuming that fetchdata.kvs returns a string[/comment]
				[comment]# We evaluate the return value thru ${} and echo it[/comment]
				echo ${ parse -r /home/pragma/fetchdata.kvs; };
			[/example]
	*/
	
	_KVS_CORESIMPLECOMMAND(parse)
	{
		QString szFileName;
		pParams->firstAsString(szFileName);

		KviFileUtils::adjustFilePath(szFileName);

		if(szFileName.isEmpty())
		{
			if(!pSwitches->find('q'))c->warning(__tr2qs("Missing file name to parse"));
			return true;
		}

		if(KviFileUtils::directoryExists(szFileName))
		{
			// FIXME: Throw away KviStr!!!
			KviStr szTmp;
			
			c->enterBlockingSection();
			
			bool bResult = KviFileDialog::askForOpenFileName(szTmp,__tr2qs("Choose a file to parse"),szFileName.latin1(),"*.kvs");

			if(!c->leaveBlockingSection())return false; // need to stop immediately
			if(!bResult)return true;

			szFileName = szTmp;

			if(szFileName.isEmpty())return true; // done
			KviFileUtils::adjustFilePath(szFileName);
		}

		QString * pszBuffer = new QString;
		if(!KviFileUtils::loadFile(szFileName,*pszBuffer,true))
		{
			delete pszBuffer;
			if(pSwitches->find('e'))
			{
				c->error(__tr2qs("Failed to load the file '%Q' for parsing"),&szFileName);
				return false;
			} else {
				if(!pSwitches->find('q'))c->warning(__tr2qs("Failed to load the file '%Q' for parsing"),&szFileName);
				return true;
			}
		}

		KviKvsScript s(szFileName,pszBuffer);

		KviKvsVariant vFileName(szFileName);
		KviKvsVariantList parms(&vFileName);
		for(KviKvsVariant * v = pParams->next();v;v = pParams->next())parms.append(v);
		parms.setAutoDelete(false);

		int iFlags = KviKvsScript::PrintErrorsAndWarnings | KviKvsScript::PreserveParams;

		KviKvsVariant * pRetVal = pSwitches->find('r') ? c->returnValue() : 0;

		if(!s.run(c->window(),&parms,pRetVal,iFlags))
		{
			if(pSwitches->find('f'))return false;
		}

		return true;
	}
};

